from unittest import TestCase
from nose.tools import (raises, assert_equal, assert_true, assert_false,
    assert_almost_equal)
from iminuit import Minuit

def assert_array_almost_equal(actual, expected):
    """
    Helper function to test if all elements of a list of lists
    are almost equal.
    Like numpy.testing.assert_array_almost_equal,
    which we can't use here because we don't
    want to depend on numpy.
    """
    for row in range(len(actual)):
        for col in range(len(actual[0])):
            assert_almost_equal(actual[row][col], expected[row][col])


class Func_Code:
    def __init__(self,varname):
        self.co_varnames=varname
        self.co_argcount=len(varname)


class Func1:
    def __init__(self):
        pass
    def __call__(self,x,y):
        return (x-2.)**2 + (y-5.)**2 + 10


class Func2:
    def __init__(self):
        self.func_code = Func_Code(['x','y'])
    def __call__(self,*arg):
        return (arg[0]-2.)**2 + (arg[1]-5.)**2 + 10


def func3(x,y):
    return 0.2*(x-2.)**2 + (y-5.)**2 + 10


def func4(x,y,z):
    return 0.2*(x-2.)**2 + 0.1*(y-5.)**2 + 0.25*(z-7.)**2 + 10


def functesthelper(f):
    m = Minuit(f, print_level=0, pedantic=False)
    m.migrad()
    val = m.values
    assert_almost_equal(val['x'], 2.)
    assert_almost_equal(val['y'], 5.)
    assert_almost_equal(m.fval, 10.)
    assert(m.matrix_accurate())
    assert(m.migrad_ok())


def test_f1():
    functesthelper(Func1())


def test_f2():
    functesthelper(Func2())


def test_f3():
    functesthelper(func3)


@raises(RuntimeError)
def test_typo():
    Minuit(func4, printlevel=0)
    #self.assertRaises(RuntimeError,Minuit,func4,printlevel=0)


def test_non_invertible():
    #making sure it doesn't crash
    def f(x,y):
        return (x*y)**2
    m = Minuit(f,pedantic=False, print_level=0)
    result = m.migrad()
    m.hesse()
    try:
        m.matrix()
        assert False #shouldn't reach here
    except RuntimeError as e:
        pass


def test_fix_param():
    m = Minuit(func4,print_level=0, pedantic=False)
    m.migrad()
    m.minos()
    val = m.values
    assert_almost_equal(val['x'],2.)
    assert_almost_equal(val['y'],5.)
    assert_almost_equal(val['z'],7.)
    err = m.errors#second derivative

    #now fix z = 10
    m = Minuit(func4,print_level=-1,y=10.,fix_y=True, pedantic=False)
    m.migrad()
    val = m.values
    assert_almost_equal(val['x'],2.)
    assert_almost_equal(val['y'],10.)
    assert_almost_equal(val['z'],7.)
    assert_almost_equal(m.fval,10.+2.5)
    free_param = m.list_of_vary_param()
    fix_param = m.list_of_fixed_param()
    assert_true('x' in free_param)
    assert_false('x' in fix_param)
    assert_true('y' in fix_param)
    assert_false('y' in free_param)
    assert_false('z' in fix_param)


def test_fitarg():
    m = Minuit(func4,print_level=-1,y=10.,fix_y=True,limit_x=(0,20.),
        pedantic=False)
    fitarg = m.fitarg
    assert_almost_equal(fitarg['y'],10.)
    assert_true(fitarg['fix_y'])
    assert_equal(fitarg['limit_x'],(0,20))
    m.migrad()

    fitarg = m.fitarg

    assert_almost_equal(fitarg['y'],10.)
    assert_almost_equal(fitarg['x'],2.,places=2)
    assert_almost_equal(fitarg['z'],7.,places=2)

    assert_true('error_y' in fitarg)
    assert_true('error_x' in fitarg)
    assert_true('error_z' in fitarg)

    assert_true(fitarg['fix_y'])
    assert_equal(fitarg['limit_x'],(0,20))


class TestErrorMatrix(TestCase):

    def setUp(self):
        self.m = Minuit(func3,print_level=0, pedantic=False)
        self.m.migrad()


    def test_error_matrix(self):
        actual = self.m.matrix()
        expected = [[5.,0.],[0.,1.]]
        assert_array_almost_equal(actual, expected)


    def test_error_matrix_correlation(self):
        actual = self.m.matrix(correlation=True)
        expected = [[1.,0.],[0.,1.]]
        assert_array_almost_equal(actual, expected)


if __name__ == '__main__':
    unittest.main()



__all__ = ['HtmlFrontend']
from IPython.core.display import display, HTML, display_html
from util import Struct
class Gradient:
    #from http://code.activestate.com/recipes/266466-html-colors-tofrom-rgb-tuples/
    @classmethod
    def color_for(cls, v, min=0., max=1., startcolor=(163,254,186),
                  stopcolor=(255,117,117)):
        c = [0]*3
        for i,sc in enumerate(startcolor):
            c[i] = round(startcolor[i] + \
                   1.0*(v-min)*(stopcolor[i]-startcolor[i])/(max-min))
        return tuple(c)

    @classmethod
    def rgb_color_for(cls,v):
        c = cls.color_for(abs(v))
        return 'rgb(%d,%d,%d)'%c

good_style = 'background-color:#92CCA6'
bad_style = 'background-color:#FF7878'
warn_style = 'background-color:#FFF79A'

def good(x,should_be):
    return good_style if x == should_be else bad_style

def caution(x,should_be):
    return good_style if x == should_be else warn_style

def fmin_style(sfmin):
    """convert sfmin to style"""
    return Struct(
        is_valid=good(sfmin.is_valid,True),
        has_valid_parameters=good(sfmin.has_valid_parameters,True),
        has_accurate_covar=good(sfmin.has_accurate_covar,True),
        has_posdef_covar=good(sfmin.has_posdef_covar,True),
        has_made_posdef_covar=good(sfmin.has_made_posdef_covar,False),
        hesse_failed=good(sfmin.hesse_failed,False),
        has_covariance=good(sfmin.has_covariance,True),
        is_above_max_edm=good(sfmin.is_above_max_edm,False),
        has_reached_call_limit=caution(sfmin.has_reached_call_limit,False),
        )


def minos_style(smerr):
    """Convert minos error to style"""
    return Struct(
        is_valid = good(smerr.is_valid,True),
        lower_valid = good(smerr.lower_valid,True),
        upper_valid = good(smerr.upper_valid,True),
        at_lower_limit = good(smerr.at_lower_limit,False),
        at_upper_limit = good(smerr.at_upper_limit,False),
        at_lower_max_fcn = good(smerr.at_lower_max_fcn,False),
        at_upper_max_fcn = good(smerr.at_upper_max_fcn,False),
        lower_new_min = good(smerr.lower_new_min,False),
        upper_new_min = good(smerr.upper_new_min,False),
        )


class HtmlFrontend:

    def print_fmin(self, sfmin, tolerance=None, ncalls=0):
        """Display FunctionMinum in html representation. 

        .. note: Would appreciate if someone would make jquery hover 
        description for each item."""
        goaledm = 0.0001*tolerance*sfmin.up
        style = fmin_style(sfmin)
        header = u"""
        <table>
            <tr>
                <td>FCN = {sfmin.fval}</td>
                <td>NFCN = {sfmin.nfcn}</td>
                <td>NCALLS = {ncalls}</td>
            </tr>
            <tr>
                <td>EDM = {sfmin.edm}</td>
                <td>GOAL EDM = {goaledm}</td>
                <td>UP = {sfmin.up}</td>
            </tr>
        </table>
        """.format(**locals())
        status = u"""
        <table>
            <tr>
                <td align="center">Valid</td>
                <td align="center">Valid Param</td>
                <td align="center">Accurate Covar</td>
                <td align="center">PosDef</td>
                <td align="center">Made PosDef</td>
            </tr>
            <tr>
                <td align="center" style="{style.is_valid}">{sfmin.is_valid!r}</td>
                <td align="center" style="{style.has_valid_parameters}">{sfmin.has_valid_parameters!r}</td>
                <td align="center" style="{style.has_accurate_covar}">{sfmin.has_accurate_covar!r}</td>
                <td align="center" style="{style.has_posdef_covar}">{sfmin.has_posdef_covar!r}</td>
                <td align="center" style="{style.has_made_posdef_covar}">{sfmin.has_made_posdef_covar!r}</td>
            </tr>
            <tr>
                <td align="center">Hesse Fail</td>
                <td align="center">HasCov</td>
                <td align="center">Above EDM</td>
                <td align="center"></td>
                <td align="center">Reach calllim</td>
            </tr>
            <tr>
                <td align="center" style="{style.hesse_failed}">{sfmin.hesse_failed!r}</td>
                <td align="center" style="{style.has_covariance}">{sfmin.has_covariance!r}</td>
                <td align="center" style="{style.is_above_max_edm}">{sfmin.is_above_max_edm!r}</td>
                <td align="center"></td>
                <td align="center" style="{style.has_reached_call_limit}">{sfmin.has_reached_call_limit!r}</td>
            </tr>
        </table>
        """.format(**locals())
        display_html(header+status,raw=True)


    def print_merror(self, vname, smerr):
        stat = 'VALID' if smerr.is_valid else 'PROBLEM'
        style = minos_style(smerr)
        to_print = """
        <span>Minos status for {vname}: <span style="{style.is_valid}">{stat}</span></span>
        <table>
            <tr>
                <td>Error</td>
                <td>{smerr.lower}</td>
                <td>{smerr.upper}</td>
            </tr>
            <tr>
                <td>Valid</td>
                <td style="{style.lower_valid}">{smerr.lower_valid}</td>
                <td style="{style.upper_valid}">{smerr.upper_valid}</td>
            </tr>
            <tr>
                <td>At Limit</td>
                <td style="{style.at_lower_limit}">{smerr.at_lower_limit}</td>
                <td style="{style.at_upper_limit}">{smerr.at_upper_limit}</td>
            </tr>
            <tr>
                <td>Max FCN</td>
                <td style="{style.at_lower_max_fcn}">{smerr.at_lower_max_fcn}</td>
                <td style="{style.at_upper_max_fcn}">{smerr.at_upper_max_fcn}</td>
            </tr>
            <tr>
                <td>New Min</td>
                <td style="{style.lower_new_min}">{smerr.lower_new_min}</td>
                <td style="{style.upper_new_min}">{smerr.upper_new_min}</td>
            </tr>
        </table>
        """.format(**locals())
        display_html(to_print, raw=True)


    def print_param(self, mps, merr=None):
        """print list of parameters"""
        #todo: how to make it right clickable to export to latex
        to_print = ""
        header = """
        <table>
            <tr>
                <td></td>
                <td>Name</td>
                <td>Value</td>
                <td>Parab Error</td>
                <td>Minos Error-</td>
                <td>Minos Error+</td>
                <td>Limit-</td>
                <td>Limit+</td>
                <td>FIXED</td>
            </tr>
        """.format(**locals())
        to_print += header
        for i,mp in enumerate(mps):
            minos_p, minos_m = (0.,0.) if merr is None or mp.name not in merr else\
                               (merr[mp.name].upper, merr[mp.name].lower)
            limit_p, limit_m = ('','') if not mp.has_limits else\
                               (mp.upper_limit, mp.lower_limit)
            fixed = 'FIXED' if mp.is_fixed else ''
            j = i+1
            content = """
            <tr>
                <td>{j}</td>
                <td>{mp.name}</td>
                <td>{mp.value:e}</td>
                <td>{mp.error:e}</td>
                <td>{minos_m:e}</td>
                <td>{minos_p:e}</td>
                <td>{limit_m!s}</td>
                <td>{limit_p!s}</td>
                <td>{fixed}</td>
            </tr>
            """.format(**locals())
            to_print += content
        to_print += """
            </table>
        """
        display_html(to_print, raw=True)


    def print_banner(self, cmd):
        #display_html('<h2>%s</h2>'%cmd, raw=True)
        pass


    def print_matrix(self, vnames, matrix, varno=None):
        to_print = """
            <table>
                <tr>
                    <td></td>
        """
        for v in vnames:
            to_print += """
            <td>
            <span style="-webkit-writing-mode:vertical-rl;-moz-writing-mode: vertical-rl;writing-mode: vertical-rl;">
            {v}
            </span>
            </td>
            """.format(**locals())
        to_print += """
                </tr>
                """
        for i,v1 in enumerate(vnames):
            to_print += """
            <tr>
                <td>{v1}</td>
            """.format(**locals())
            for j,v2 in enumerate(vnames):
                val = matrix[i][j]
                color = Gradient.rgb_color_for(val)
                to_print += """
                <td style="background-color:{color}">
                {val:3.2f}
                </td>
                """.format(**locals())
            to_print += """
            </tr>
            """
        to_print += '</table>'
        display_html(to_print, raw=True)


    def print_hline(self):
        display_html('<hr>', raw=True)


class HTMLWrapper:
    def __init__(self, txt):
        self.data = txt
    def _repr_html_(self):
        return txt



#profile(self, vname, bins=100, bound=2, args=None, subtract_min=False)
def draw_profile(self, vname, bins=100, bound=2, args=None, subtract_min=False):
    from matplotlib import pyplot as plt
    import numpy as np
    x, y = self.profile(vname, bins, bound, args, subtract_min)
    x = np.array(x)
    y = np.array(y)

    plt.plot(x,y)
    plt.grid(True)
    plt.xlabel(vname)
    plt.ylabel('FCN')

    minpos = np.argmin(y)

    ymin = y[minpos]
    tmpy = y - ymin
    #now scan for minpos to the right until greater than one
    up = self.up
    righty = np.power(tmpy[minpos:] - up, 2)
    #print righty
    right_min = np.argmin(righty)
    rightpos = right_min + minpos
    lefty = np.power((tmpy[:minpos] - up), 2)
    left_min = np.argmin(lefty)
    leftpos = left_min
    #print leftpos, rightpos
    le = x[minpos] - x[leftpos]
    re = x[rightpos] - x[minpos]
    #print (le, re)
    p = plt.axvspan(x[leftpos], x[rightpos], facecolor='g', alpha=0.5)
    plt.figtext(0.5, 0.5, '%s = %7.3e ( -%7.3e , +%7.3e)' % (vname, x[minpos], le, re), ha='center')
    return x,y

def draw_contour(self, x, y, bins=20, bound=2, args=None, show_sigma=False):
#def draw_contour(self, var1, var2, bins=12, bound1=None, bound2=None, lh=True):
    from matplotlib import pyplot as plt
    import numpy as np
    vx, vy, vz = self.contour(x, y, bins, bound, args, subtract_min=True)
    #x1s, x2s, y = val_contour2d(fit, m, var1, var2, bins=bins,
    #                            bound1=bound1, bound2=bound2)
    #y -= np.min(y)

    v = [self.up*(i**2) for i in range(1,4)]

    CS = plt.contour(vx, vy, vz, v, colors=['b', 'k', 'r'])
    if not show_sigma:
        plt.clabel(CS, v)
    else:
        tmp = dict((vv,r'%i $\sigma$'%(i+1)) for i,vv in enumerate(v))
        plt.clabel(CS, v, fmt=tmp, fontsize=16)
    plt.xlabel(x)
    plt.ylabel(y)
    plt.axhline(self.values[y], color='k', ls='--')
    plt.axvline(self.values[x], color='k', ls='--')
    plt.grid(True)
    return vx, vy, vz



import warnings

class IMinuitWarning(RuntimeWarning):
    pass

class InitialParamWarning(IMinuitWarning):
    pass

class HesseFailedWarning(IMinuitWarning):
    pass

warnings.simplefilter('always', InitialParamWarning, append=True);



class ConsoleFrontend:
    """Minuit console front end class
    This class print stuff directly via print
    #TODO: add some color
    """
    def print_fmin(self,sfmin, tolerance=None, ncalls = 0):
        """display function minimum information
        for FunctionMinimumStruct *sfmin*.
        It contains various migrad status.
        """
        fmin = sfmin
        goaledm = 0.0001*tolerance*fmin.up if tolerance is not None else ''
        #despite what the doc said the code is actually 1e-4
        #http://wwwasdoc.web.cern.ch/wwwasdoc/hbook_html3/node125.html
        flatlocal = dict(locals().items()+fmin.__dict__.items())
        info1 = 'fval = %(fval)r | nfcn = %(nfcn)r | ncalls = %(ncalls)r\n'%\
                flatlocal
        info2 = 'edm = %(edm)r (Goal: %(goaledm)r) | up = %(up)r\n'%flatlocal
        header1 = '|' + (' %14s |'*5)%(
                    'Valid',
                    'Valid Param',
                    'Accurate Covar',
                    'Posdef',
                    'Made Posdef')+'\n'
        hline = '-'*len(header1)+'\n'
        status1 = '|' + (' %14r |'*5)%(
                    fmin.is_valid,
                    fmin.has_valid_parameters,
                    fmin.has_accurate_covar,
                    fmin.has_posdef_covar,
                    fmin.has_made_posdef_covar)+'\n'
        header2 = '|' + (' %14s |'*5)%(
                    'Hesse Fail',
                    'Has Cov',
                    'Above EDM',
                    '',
                    'Reach calllim')+'\n'
        status2 = '|' + (' %14r |'*5)%(
                    fmin.hesse_failed,
                    fmin.has_covariance,
                    fmin.is_above_max_edm,
                    '',
                    fmin.has_reached_call_limit)+'\n'

        print hline + info1 + info2 +\
            hline + header1 + hline + status1 +\
            hline + header2 + hline+ status2 +\
            hline


    def print_merror(self,vname,smerr):
        """print minos error for varname"""
        stat = 'VALID' if smerr.is_valid else 'PROBLEM'

        summary = 'Minos Status for %s: %s\n'%\
                (vname,stat)

        error = '| {:^15s} | {: >12g} | {: >12g} |\n'.format(
                    'Error',
                    smerr.
                    lower,
                    smerr.upper)
        valid = '| {:^15s} | {:^12s} | {:^12s} |\n'.format(
                    'Valid',
                    str(smerr.lower_valid),
                    str(smerr.upper_valid))
        at_limit='| {:^15s} | {:^12s} | {:^12s} |\n'.format(
                    'At Limit',
                    str(smerr.at_lower_limit),
                    str(smerr.at_upper_limit))
        max_fcn='| {:^15s} | {:^12s} | {:^12s} |\n'.format(
                    'Max FCN',
                    str(smerr.at_lower_max_fcn),
                    str(smerr.at_upper_max_fcn))
        new_min='| {:^15s} | {:^12s} | {:^12s} |\n'.format(
                    'New Min',
                    str(smerr.lower_new_min),
                    str(smerr.upper_new_min))
        hline = '-'*len(error)+'\n'
        print hline +\
              summary +\
              hline +\
              error +\
              valid +\
              at_limit +\
              max_fcn +\
              new_min +\
              hline


    def print_param(self, mps, merr=None):
        """Print parameter states

        Arguments:

            *mps*: list of MinuitParameter struct

            *merr*: dictionary of vname->minos error struct
        """
        def lud(m,k,d):
            #lookup map with default
            return m[k] if k in m else d
        merr = {} if merr is None else merr
        vnames = [mp.name for mp in mps]
        maxlength = max([len(x) for x in vnames])
        maxlength = max(5,maxlength)

        header = (('| {:^4s} | {:^%ds} | {:^8s} | {:^8s} | {:^8s} |'
                  ' {:^8s} | {:^8s} | {:^8s} | {:^8s} |\n')%maxlength).format(
                    '','Name', 'Value','Para Err',
                    "Err-","Err+","Limit-","Limit+"," ")
        hline = '-'*len(header)+'\n'
        linefmt = ('| {:>4d} | {:>%ds} = {:<8s} | {:<8s} | {:<8s} |'
                  ' {:<8s} | {:<8s} | {:<8s} | {:^8s} |\n')%maxlength
        nfmt = '{:< 8.4G}'
        nformat = nfmt.format
        blank = ' '*8

        ret = hline+header+hline
        for i,(v,mp) in enumerate(zip(vnames,mps)):
            tmp = [i,v]

            tmp.append(nfmt.format(mp.value))
            tmp.append(nfmt.format(mp.error))

            tmp.append(nformat(merr[v].lower) if v in merr else blank)
            tmp.append(nformat(merr[v].upper) if v in merr else blank)

            tmp.append(nformat(mp.lower_limit) if mp.has_limits else blank)
            tmp.append(nformat(mp.upper_limit) if mp.has_limits else blank)

            tmp.append(
                'FIXED' if mp.is_fixed else 'CONST' if mp.is_const else '')

            line = linefmt.format(*tmp)
            ret+=line
        ret+=hline
        print ret


    def print_banner(self, cmd):
        """show banner of command"""
        ret = '*'*50+'\n'
        ret += '*{:^48}*'.format(cmd)+'\n'
        ret += '*'*50+'\n'
        print ret


    def print_matrix(self,vnames, matrix, varno=None):
        """TODO: properly implement this"""
        print vnames
        print matrix
        maxlen = max(len(v) for v in vnames)
        narg = len(matrix)
        vfmt = '%%%ds'%maxlen
        vblank = ' '*maxlen
        fmt = '%3.2f '#4char
        dfmt = '%4d '
        tmp = ''
        header = vblank+' '*4+'  | '+(dfmt*narg)%tuple(range(narg))+'\n'
        blank_line = '-'*len(header)+'\n'
        tmp += header + blank_line
        for i, (v, row) in enumerate(zip(vnames,matrix)):
            fmt = '%3.2f '*narg
            head = (vfmt+' %4d | ')%(v, i)
            content = (fmt)%tuple(row)+'\n'
            tmp += head + content
        tmp += blank_line
        print tmp


    def print_hline(self):
        print '*'*70



# -*- coding: utf8 -*-
from distutils.core import setup, Extension
import distutils.util
import subprocess
from os.path import dirname, join
from glob import glob
from distutils.ccompiler import new_compiler

#static link
cwd = dirname(__file__)
minuit_src = glob(join(cwd,'Minuit/src/*.cpp'))
minuit_header = join(cwd,'Minuit')

libiminuit = Extension('iminuit._libiminuit',
                    sources = ['iminuit/_libiminuit.cpp'] + minuit_src,
                    include_dirs= [minuit_header],
                    libraries = [],
                    extra_compile_args = ['-Wno-write-strings', '-Wno-sign-compare'],
                    extra_link_args = [])

execfile('iminuit/info.py')

setup (
    name = 'iminuit',
    version = __version__,
    description = 'Interactive Minimization Tools based on MINUIT',
    long_description=''.join(open('README.rst').readlines()[4:]),
    author='Piti Ongmongkolkul',
    author_email='piti118@gmail.com',
    url='https://piti118.github.com/iminuit',
    download_url='http://pypi.python.org/packages/source/i/'
            'iminuit/iminuit-%s.tar.gz' % __version__,
    package_dir = {'iminuit': 'iminuit'},
    packages = ['iminuit'],
    ext_modules = [libiminuit],
    classifiers=[
        "Programming Language :: Python",
        'Topic :: Scientific/Engineering :: Physics',
        'Topic :: Scientific/Engineering :: Mathematics',
        'Intended Audience :: Science/Research',
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: GNU General Public License (GPL)',
        'License :: OSI Approved :: MIT License'
    ]
)
